"""Provides a layer that presents file system in a table-like fashion, to simplify bulk loading into data frames.

The goal is not to replace advanced tools such as hive metastore, athena/wrangler, spark.
Rather, we aim to cover some corner cases where those tools are not set up or cannot be applied.
In the case of tables being stored with:
 * `<key>=<value>` format of partition columns,
 * parquet files having flat schema,
 * filenames bearing no information,
this module should bring no advantage over general tools.

Primary API consists of:
 - `read_partitioned_table`,
 - `write_object`.

The conceptual separation of submodules consists of:
 - `column_parser` captures how the filesystem path is translated into table *partition* columns,
 - `query` handles filtering of the partition columns,
 - `deser` deals with conversion of the raw bytes into various dataframes or other user-prefered structures,
 - `partition_discovery` -- controls the main logic of iterative crawling and filter invocation.

At the moment, interactions with an underlying filesystem is handled via the `fsspec` lib -- which is externally
exposed so that the user can fully configure it at will.

A listing of a filesystem, if done without care, can result in a needless delays and extra costs.
Some care is paid to optimisation here, mostly in the `partition_discovery` module. In brief, the user is
advised to:
1. use `column_parser`'s generative functionality, if a range of some column is small -- this is invoked *before*
   the call to list file system's content,
2. use `query`'s composition of queries, as opposed to compose on their own, and use as few columns as possible
   in the individual subqueries (atomic statements) -- this is because subqueries are evaluated eagerly, once all
   columns are available.

As with every distributed query system, one should understand there are no transactional guarantees. In particular:
 - if, between partition discovery and reading, a file is deleted, the query crashes with FileNotFound
 - if, between partition discovery and reading, a file is added, the query does not process it.
"""

from __future__ import annotations

import io
import logging
import shutil
import warnings
from typing import Optional, Union

import pandas as pd
from fsspec.spec import AbstractFileSystem

from fsql import get_url_and_fs
from fsql.column_parser import AUTO_PARSER, ColumnParser
from fsql.deser import PANDAS_READER, DataReader
from fsql.partition import Partition
from fsql.partition_discovery import discover_partitions
from fsql.query import Query

logger = logging.getLogger(__name__)


def read_s3_table(
    url: str,
    query: Query,
    column_parser: ColumnParser = AUTO_PARSER,
    data_reader: DataReader = PANDAS_READER,
):
    """Old deprecated name, use `read_partitioned_table` instead"""
    warnings.warn("Function `read_s3_table` is deprecated, use `read_partitioned_table` instead", DeprecationWarning)
    return read_partitioned_table(url, query, column_parser, data_reader)


def read_partitioned_table(
    url: str,
    query: Query,
    column_parser: ColumnParser = AUTO_PARSER,
    data_reader: DataReader = PANDAS_READER,
    fs: Optional[AbstractFileSystem] = None,
):
    """Reads a table rooted at `url`, with partition columns described in `column_parser` and filtered via `query`.

    The default values assume `colName1=val/colName2=val` format of the path, and pandas data frame as output format.
    There is no default query -- for reading all partitions, user is supposed to use the always-true query from the
    `query` module.

    Note that the provided DataReaders launch a ThreadPoolExecutor when downloading individual files, to speed up I/O.

    If `fs` is not provided, a default one is constructed from the url. The instance is then used for all `ls`
    and `open` operations.
    """
    url_suff, fs_default = get_url_and_fs(url)
    fs = fs if fs else fs_default
    # we are relying on the invariant that directory urls end with '/'
    # we thus need to check whether the user is not fooling us
    if not url_suff.endswith("/"):
        if fs.isdir(url_suff):
            url_suff += "/"

    root_partition = Partition(url_suff, {})
    logging.debug(f"partition discovery starting. Url: {url_suff}, Query: {query}")
    partitions = discover_partitions(query, column_parser, root_partition, fs)
    partitions = list(partitions)
    logging.debug(f"partitions are {partitions}")
    logging.debug(f"data fetch starting. Url: {url}, Query: {query}")
    return data_reader.read_and_concat(partitions, fs)


def write_object(
    url: str,
    data: Union[pd.DataFrame, io.StringIO, io.BytesIO],
    format: Optional[str] = None,
    format_options: Optional[dict[str, str]] = None,
    fs: Optional[AbstractFileSystem] = None,
):
    """Minimalistic function to write an object to a designated location.

    * Does not support any table-like semantics -- partition appends or multi-partition inserts,
    * for `io` objects, `seek(0)` and `shutil` write is the behaviour; `format` argument must not be specified,
    * for Data Frames converts to the specified format (only parquet+fastparquet for now), and then it is written
      using the file-like object from fsspec and dataframe's native write.

    At the moment, only parquet and csv as a format are supported. The only format options are 'engine' with values
    'fastparquet' and 'pyarrow' (these refer only to parquet format).

    For larger data frames or table-like semantics, use rather endpoint (TODO).

    If `fs` is not provided, a default one is constructed from the url. The instance is then used for an `open`
    operation.
    """

    url_suff, fs_default = get_url_and_fs(url)
    fs = fs if fs else fs_default
    format_options = format_options if format_options else {}

    if isinstance(data, pd.DataFrame):
        if format == "parquet" or format is None:
            engine = format_options.get("engine", "fastparquet")
            if engine == "fastparquet":
                data.to_parquet(url_suff, engine=engine, open_with=fs.open)
            elif engine == "pyarrow":
                with fs.open(url_suff, "wb") as fd:
                    data.to_parquet(fd, engine=engine)
            else:
                raise ValueError(f"unsupported engine for dataframe writing: {engine}")
        elif format == "csv":
            with fs.open(url_suff, "wb") as fd:
                data.to_csv(fd)
        else:
            raise ValueError(f"unsupported format for dataframe writing: {format}")
    elif isinstance(data, io.StringIO) or isinstance(data, io.BytesIO):
        if format:
            raise ValueError(f"cannot specify format when data is a buffer. Provided format: {format}")
        data.seek(0)
        mode = "wb" if isinstance(data, io.BytesIO) else "w"
        with fs.open(url_suff, mode) as fd:
            shutil.copyfileobj(data, fd)



"""Handles conversion of raw bytes + parsed partition columns into formats such as data frames.

The main interface `DataReader` with entrypoint method `read_and_concat` gives the contract of
the operation.  There are the following default instances:
 - PandasReader, which parses to pandas data frames and adds the partition columns as df column,
   and concatenates all the individual sub-frames into a single one
 - EnumeratedDictReader, which reads the data as a dictionary, adds the partition columns as dict
   keys, and concatenates into an enumerated dict (order being the alphabetic of underlying files)

All these autodetect the input format from the suffix of the key. If this is desired to be
overridden with a fixed format, user should instantiate with the desired InputFormat.

If the user desires to implement another format, such as Dask, the `DataReader` interface is the part of
the public API. The main methods are `read_single` (which is executed with a thread pool executor, as
we want to parallelize cloud FS fetches for an acceptable performance) and `concat`. It may happen that
this approach is undesirable due to too-large intermediate representation -- in this case, user should
opt for the lazy approach (such as in Dask), and don't materialize inside neither the `read_single` nor
`concat` methods.

Existing readers such as PandasReader allow customisation via passing through any kwargs to the underlying
pandas read method.

The user should *not* bake in any specific business logic in here -- a more prefered approach is to
return an object such as (lazy) data frame as early as possible, and apply any transformations later on.
"""
from __future__ import annotations

import json
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor
from enum import Enum, auto, unique
from functools import partial
from typing import Any

import pandas as pd
from fsspec.spec import AbstractFileSystem

from fsql import assert_exhaustive_enum
from fsql.partition import Partition

logger = logging.getLogger(__name__)


@unique
class InputFormat(Enum):
    # NOTE possibly remove this enum as it requires more exhaustive match, and instead use Union[InputFormat, Auto]
    # TODO behaviour for gzipped etc files. Either two dim enum, or concat in here
    AUTO = auto()
    PARQUET = auto()
    JSON = auto()
    CSV = auto()
    XLSX = auto()

    @classmethod
    def from_url(cls, url: str):
        return {
            "json": InputFormat.JSON,
            "parquet": InputFormat.PARQUET,
            "csv": InputFormat.CSV,
            "xlsx": InputFormat.XLSX,
        }[url]


class DataReader(ABC):
    def __init__(self, input_format: InputFormat = InputFormat.AUTO):
        self.input_format = input_format

    def detect_format(self, url: str) -> InputFormat:
        if self.input_format != InputFormat.AUTO:
            return self.input_format
        else:
            return InputFormat.from_url(url.split(".")[-1])

    @abstractmethod
    def read_single(self, partition: Partition, fs: AbstractFileSystem) -> Any:  # TODO generics on the Any
        raise NotImplementedError("abc")

    @abstractmethod
    def concat(self, data: Iterable[Any]) -> Any:  # TODO generics on the Any
        raise NotImplementedError("abc")

    def read_and_concat(
        self, partitions: Iterable[Partition], fs: AbstractFileSystem
    ) -> Any:  # TODO generics on the Any
        with ThreadPoolExecutor(max_workers=32) as tpe:  # TODO configurable
            data = tpe.map(partial(self.read_single, fs=fs), partitions)
        return self.concat(data)


class PandasReader(DataReader):
    """Wraps various pandas read methods (parquet, json, csv, excel) into a single interface.
    Behaviour can be customised via passing any kwargs to the constructor.
    """

    def __init__(self, input_format=InputFormat.AUTO, **pdread_kwargs):
        super().__init__(input_format=input_format)
        self.pdread_user_kwargs = pdread_kwargs
        self.pdread_default_kwargs = defaultdict(dict)
        self.pdread_default_kwargs[InputFormat.PARQUET] = {
            "engine": "fastparquet",
        }
        self.pdread_default_kwargs[InputFormat.JSON] = {
            "lines": "true",
        }
        self.pdread_default_kwargs[InputFormat.XLSX] = {
            "engine": "openpyxl",
        }

    def read_single(self, partition: Partition, fs: AbstractFileSystem) -> pd.DataFrame:
        logger.debug(f"read dataframe for partition {partition}")
        input_format = self.detect_format(partition.url)
        logger.debug(f"format detected for partition {input_format} <- {partition}")
        if input_format is InputFormat.PARQUET:
            reader = pd.read_parquet
        elif input_format is InputFormat.JSON:
            reader = pd.read_json
        elif input_format is InputFormat.CSV:
            reader = pd.read_csv
        elif input_format is InputFormat.XLSX:
            reader = pd.read_excel
        elif input_format is InputFormat.AUTO:
            raise ValueError(f"partition had format detected as auto -> invalid state. Partition: {partition}")
        else:
            assert_exhaustive_enum(input_format)

        pdread_kwargs = {**self.pdread_default_kwargs[input_format], **self.pdread_user_kwargs}
        logger.debug(f"reader kwargs {pdread_kwargs} for partition {partition}")
        try:
            with fs.open(partition.url, "rb") as fd:
                df = reader(fd, **pdread_kwargs)
        except FileNotFoundError as e:
            logger.warning(f"file {partition} reading exception {type(e)}, attempting cache invalidation and reread")
            fs.invalidate_cache()
            with fs.open(partition.url, "rb") as fd:
                df = reader(fd, **pdread_kwargs)

        for key, value in partition.columns.items():
            df[key] = value
        return df

    def concat(self, data: Iterable[Any]) -> Any:  # TODO generics on the Any
        return pd.concat(data)


PANDAS_READER = PandasReader()


class EnumeratedDictReader(DataReader):
    def read_single(self, partition: Partition, fs: AbstractFileSystem) -> dict:
        logger.debug(f"read single for partition {partition}")
        input_format = self.detect_format(partition.url)
        if input_format != InputFormat.JSON:
            raise ValueError(f"EnumeratedDictReader supports only json, not {input_format}. Partition is {partition}")

        def read_json(url):
            with fs.open(url, "r") as fd:
                return json.load(fd)

        try:
            base = read_json(partition.url)
        except FileNotFoundError as e:
            logger.warning(f"file {partition} reading exception {type(e)}, attempting cache invalidation and reread")
            fs.invalidate_cache()
            base = read_json(partition.url)
        return {**base, **partition.columns}

    def concat(self, data: Iterable[Any]) -> Any:  # TODO generics on the Any
        return dict(enumerate(data))


ENUMERATED_DICT_READER = EnumeratedDictReader()



"""Captures how the filesystem path is translated into table *partition* columns.

There are two standard formats: colName1=val/colName2=val (e.g., year=2022/month=10/...),
which we call Auto, and val/val (e.g., 2022/10) which we call FixedColumns. More arcane
approaches can be handled by implementing the interface `ColumnParser`, which is part of this
module's public API.

In more formal words, `ColumnParser` describes the _grammar_ of the table's partition columns, because,
for optimisation reasons, it also has `generate` method -- be sure to use it whenever possible,
so that the `partition_discovery` module can exploit this.

The standard usage of this module should consist of constructing instances of the AutoParser or
FixedColumns parser using the `from_str` methods -- see their individual documentations.
"""

# TODO
# This may better be named 'Table' or 'TableContract', or even PartitionsGrammar. In that case, the __call__ method
# would rather have an explicit parse name.
# More importantly, the whole class hierarchy in here is wrong. We have a single ColumnParser class with two
# implementations. But there are actually four separate concerns:
# - grammar of the column partitions (includes both parser and generator)
# - pre-specifying some partition values (ie, providing input for generator)
# - exhaustiveness of parsing (early stopping or detection thereof)
# - whether filenames are parsed
# And covering that with just one hierarchy is not a good idea. Instead, we should make this more fine-grained,
# with a number of pre-configured classes to choose from (and possibly configure them further).
# This will get us rid of that weird calling of init of an abstract class.

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from fsql.partition import Partition

logger = logging.getLogger(__name__)


@dataclass
class PartitionGrammar:  # this is a quite bad name
    name: str
    values: Optional[list[str]]


class ColumnParser(ABC):
    @abstractmethod
    def __call__(self, dirname: str) -> tuple[str, str]:
        raise NotImplementedError("abc")

    @abstractmethod
    def tail(self, partition: Partition) -> ColumnParser:
        raise NotImplementedError("abc")

    @abstractmethod
    def generate(self) -> Optional[list[str]]:
        raise NotImplementedError("abc")

    @abstractmethod
    def parses_filenames(self) -> bool:
        raise NotImplementedError("abc")

    @abstractmethod
    def is_terminal_level(self) -> bool:
        raise NotImplementedError("abc")

    @classmethod
    def from_str(cls, path_description: str):
        """Example: col1/col2=v1/col3=[v4,v5,v6]/colFname"""
        # NOTE this whole method better be replaced with a proper parser
        def process_single_partition(partition_desc: str):
            eq_split = partition_desc.split("=")
            if len(eq_split) == 1:
                return PartitionGrammar(eq_split[0], None)
            else:
                if eq_split[1][0] == "[":
                    return PartitionGrammar(eq_split[0], eq_split[1][1:-1].split(","))
                else:
                    return PartitionGrammar(eq_split[0], [eq_split[1]])

        # not sure how to correctly handle this -- I don't want to declare the __init__ here abstract...
        return cls([process_single_partition(e) for e in path_description.split("/")])  # type: ignore


class AutoParser(ColumnParser):
    def __init__(self, partition_grammars: Optional[list[PartitionGrammar]] = None):
        self.partitions = partition_grammars

    def __call__(self, dirname: str) -> tuple[str, str]:
        key, value = dirname.strip("/").split("=", 1)
        return key, value  # we don't return directly due to mypy not understanding split(_, 1)

    def tail(self, partition: Partition) -> ColumnParser:
        if not self.partitions:
            return self
        else:
            # TODO performance issue
            # ideally, the pop call would return an existing instance, which would be pre-created...
            return AutoParser(self.partitions[1:])

    def parses_filenames(self) -> bool:
        return False

    def is_terminal_level(self) -> bool:
        # NOTE a quirk -- if partitions not provided, we read files at every stage of crawling, and thus
        # can't guarantee all of them containing the same number of columns. Fixing that would require
        # first iterating through all of the discovered partitions and filtering for max length only.
        # In the same vein, we don't guarantee even that the columns are the same for every partition.
        # For that, however, the best we could do is crash in case of inconsistency...
        if not self.partitions:
            return True
        else:
            return len(self.partitions) == 0

    def generate(self) -> Optional[list[str]]:
        if self.partitions and self.partitions[0].values:
            return [f"{self.partitions[0].name}={value}" for value in self.partitions[0].values]
        else:
            return None


class FixedColumnsParser(ColumnParser):
    def __init__(self, partition_grammars: list[PartitionGrammar]):
        self.partitions = partition_grammars

    def __call__(self, dirname: str) -> tuple[str, str]:
        return (self.partitions[0].name, dirname.strip("/"))

    def tail(self, partition: Partition) -> ColumnParser:
        # TODO performance issue
        # ideally, the pop call would return an existing instance, which would be pre-created...
        return FixedColumnsParser(self.partitions[1:])

    def parses_filenames(self) -> bool:
        return True

    def is_terminal_level(self) -> bool:
        return len(self.partitions) == 1

    def generate(self) -> Optional[list[str]]:
        if not self.partitions:
            raise ValueError("no partitions remaining")
        if self.partitions[0].values:
            return self.partitions[0].values
        else:
            return None


AUTO_PARSER = AutoParser()



"""Internal dataclass module."""

from __future__ import annotations

from copy import copy
from dataclasses import dataclass
from typing import Optional


@dataclass
class Partition:
    url: str
    columns: dict[str, str]

    def expand_by(self, url_suffix: str, key_val: Optional[tuple[str, str]]) -> Partition:
        columns_ext = copy(self.columns)
        if key_val:
            if key_val[0] in columns_ext:
                raise ValueError(f"duplicate key inserted: {key_val[0]}")
            columns_ext[key_val[0]] = key_val[1]
        return Partition(self.url + url_suffix, columns_ext)



"""Controls the main logic of iterative crawling and filter invocation.

This is the module that ties `query` and `column_parser` together. It is, at the moment,
least-well designed and most fragile. One has to be careful when extending this, to not
cause severe malperformance.

This module is not principally part of the public API, we don't expect any behaviour here
to be customisable.
"""

# TODO improve design, upgrade the doc

from __future__ import annotations

import logging
from collections.abc import Iterable
from dataclasses import dataclass
from itertools import chain, groupby

from fsspec.spec import AbstractFileSystem

from fsql.column_parser import ColumnParser
from fsql.partition import Partition
from fsql.query import Query

logger = logging.getLogger(__name__)
flat_map = lambda f, xs: (y for ys in xs for y in f(ys))  # noqa: E731


@dataclass
class DirectoryListing:
    files: Iterable[str]
    directories: Iterable[str]


def list_directory(url: str, fs: AbstractFileSystem) -> DirectoryListing:
    listing_raw = fs.ls(url, detail=True)
    logger.debug(f"url {url} listed to {len(listing_raw)} elements")
    selector = lambda e: e["type"]  # noqa: E731
    extractor = lambda l: sorted(list(map(lambda e: e["name"].split("/")[-1], l)))  # noqa: E731
    # we need the extractor to materialise to list, because the groupby iterator is shared
    listing_grouped = {k: extractor(v) for k, v in groupby(sorted(listing_raw, key=selector), selector)}
    listing = DirectoryListing(files=listing_grouped.get("file", []), directories=listing_grouped.get("directory", []))
    logger.debug(f"url {url} listed to {listing}")
    return listing


def discover_partitions(
    query: Query, column_parser: ColumnParser, partition: Partition, fs: AbstractFileSystem
) -> Iterable[Partition]:
    # NOTE expose in the query the option to look at last item only, do the expand by as a flat map

    # NOTE this whole thing is quite hacky. Before querying fs, we check whether the user has prescribed the
    # values for columns at this stage exactly. The weird thing is that we still parse and query those columns later,
    # but that may actually be upside. Also, the separation between files and directories does feel a bit artificial
    # here
    logger.debug(f"partition discovery with query {query} and partition {partition}")
    generated_partitions = column_parser.generate()
    if generated_partitions:
        if column_parser.is_terminal_level():
            listing = DirectoryListing(files=generated_partitions, directories=[])
        else:
            listing = DirectoryListing(files=[], directories=[partition + "/" for partition in generated_partitions])
    else:
        if not partition.url.endswith("/"):
            partition.url += "/"  # required due to fsspec.ls not appending '/' to listed directories
        listing = list_directory(partition.url, fs)

    subdir_partitions = (partition.expand_by(item, column_parser(item)) for item in listing.directories)
    subdir_partitions_flt = filter(lambda partition: query.eval_available(partition.columns), subdir_partitions)
    # NOTE paralellisation opportunity! We don't need this to be sequential
    subdir_partitions_exp = flat_map(
        lambda partition: discover_partitions(query, column_parser.tail(partition), partition, fs),
        subdir_partitions_flt,
    )

    file_partitions_flt: Iterable[Partition] = iter(())
    if column_parser.is_terminal_level():
        if column_parser.parses_filenames():
            # TODO allow filename filtering in auto column extraction as well...
            # TODO distinguish for the fixed parser a column for the fnames...
            file_partitions = (partition.expand_by(item, column_parser(item)) for item in listing.files)
            file_partitions_flt = filter(lambda partition: query.eval_all(partition.columns), file_partitions)
        else:
            if not query.eval_all(partition.columns):
                file_partitions_flt = []
            else:
                file_partitions_flt = (partition.expand_by(item, None) for item in listing.files)

    return chain(subdir_partitions_exp, file_partitions_flt)



"""fsql"""
from __future__ import annotations

import os
from typing import Any, NoReturn

import fsspec
from fsspec.spec import AbstractFileSystem

__all__: list[str] = []


def get_url_and_fs(url: str) -> tuple[str, AbstractFileSystem]:
    """This function standardizes url->protocol derivation, and allows for fs-specific parameter passing.

    In particular, we need it to allow minio-in-place-of-s3 functionality, which is not supported on its
    own using environment variables in vanilla fsspec.
    """

    # TODO extend signature to make it more meaningufl, that is:
    # - capture the os dependency somehow, eg by having a dict param and passing os.environ to it
    # - add the user-provided FS here, to validate protocolar compatibility
    # TODO org improvements -- better location for this, and change name and return type to better capture that
    # the resource location and file system are inherently bound

    fs_key, url_suff = url.split(":/", 1)

    env2dict = lambda mapping: {val: os.environ[key] for key, val in mapping.items() if key in os.environ}  # noqa: E731

    if fs_key == "s3":
        # boto itself supports only access key and secret key via env
        # TODO this is quite dumb, especially the max pool connections -- a more intelligent way of config is desired
        # the right way forward is probably some FsFactory singleton which reads:
        # - env variables in case where it makes sense (AWS ids)
        # - config files with rich options that allow generic passthrough
        # - defaults as in the case of max_pool_connections
        # note the programmatic override can already be done by passing Fs instance to api methods
        configurable_keys_l1 = {
            "AWS_ACCESS_KEY_ID": "key",
            "AWS_SECRET_ACCESS_KEY": "secret",
            "AWS_SESSION_TOKEN": "token",
        }
        l1kwargs: dict[str, Any] = env2dict(configurable_keys_l1)  # mypy
        configurable_keys_l2 = {
            "AWS_ENDPOINT_URL": "endpoint_url",
            "AWS_REGION_NAME": "region_name",
        }
        l2kwargs = env2dict(configurable_keys_l2)
        # in case of a lot of small files and local tests, we tend to exhaust the conn pool quickly and spawning warns
        l1kwargs["config_kwargs"] = {"max_pool_connections": 25}
        l1kwargs["client_kwargs"] = l2kwargs

        fs = fsspec.filesystem("s3", **l1kwargs)
    else:
        fs = fsspec.filesystem(fs_key)

    return url_suff, fs


def assert_exhaustive_enum(x: NoReturn) -> NoReturn:
    """Python does not sux! Call this function at the end of if-else matching of Enum, to have mypy
    ensure for you that all values are considered! Cf https://tech.preferred.jp/en/blog/python-exhaustive-union-match/
    """
    raise AssertionError(f"enum matching not exhaustive: {x}")



"""Handles filtering of the partitions prior to being read, based on values of partition columns.

TODO Document more:
 - factory-based approach
 - partial query evaluation
"""
from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import Callable, Optional

# TODO
# - query should impl some caching mechanism for partitions, to not re-eval over and over
# - query could impl an optional "generate" method and have an ColEq factory method,
#   to speed up discovery with conditions like 'day=13'. This requires auto column parser
#   to allow for column order specs


class Query(ABC):
    @abstractmethod
    def eval_all(self, columns: dict[str, str]) -> bool:
        raise NotImplementedError("abc")

    @abstractmethod
    def eval_available(self, columns: dict[str, str]) -> bool:
        raise NotImplementedError("abc")


class BooleanOperatorQuery(Query):
    def __init__(self, left: Query, right: Query, operator: Callable[[bool, bool], bool]):
        self.left = left
        self.right = right
        self.operator = operator

    def eval_all(self, columns: dict[str, str]) -> bool:
        return self.operator(self.left.eval_all(columns), self.right.eval_all(columns))

    def eval_available(self, columns: dict[str, str]) -> bool:
        return self.operator(self.left.eval_available(columns), self.right.eval_available(columns))


class AtomicQuery(Query):
    def __init__(self, f: Callable, columns: Optional[set[str]] = None):
        self.f = f
        if columns:  # TODO this is to impl the Q_EQ, feels hacky
            self.columns = columns
        else:
            args = inspect.getfullargspec(f).args
            self.columns = set(args)

    def eval_all(self, columns: dict[str, str]) -> bool:
        if not self.columns.issubset(columns.keys()):
            return False
        return self.f(**{key: val for key, val in columns.items() if key in self.columns})

    def eval_available(self, columns: dict[str, str]) -> bool:
        if not self.columns.issubset(columns.keys()):
            return True
        return self.f(**{key: val for key, val in columns.items() if key in self.columns})


# NOTE for some reason, the following lines were not understood by mypyc
# thus we reimplement via usual functions
# Q_AND = lambda l, r: BooleanOperatorQuery(l, r, lambda lr, rr: lr and rr)  # noqa: E731
# Q_OR = lambda l, r: BooleanOperatorQuery(l, r, lambda lr, rr: lr or rr)  # noqa: E731


def Q_AND(l: Query, r: Query) -> Query:  # noqa: E741
    return BooleanOperatorQuery(l, r, lambda lr, rr: lr and rr)


def Q_OR(l: Query, r: Query) -> Query:  # noqa: E741
    return BooleanOperatorQuery(l, r, lambda lr, rr: lr or rr)


def Q_EQ(column: str, value: str):
    def f(**kwargs):  # TODO how to define f(`column`) ? That would simplify AtomicQuery then
        return kwargs[column] == value

    return AtomicQuery(f, set([column]))


class ConstantQuery(Query):
    def __init__(self, value: bool):
        self.constant = value

    def eval_all(self, columns: dict[str, str]) -> bool:
        return self.constant

    def eval_available(self, columns: dict[str, str]) -> bool:
        return self.constant


Q_TRUE = ConstantQuery(True)
Q_FALSE = ConstantQuery(False)



"""Implements handy utilities for querying date ranges:
- DateRangeQuery, if you need to combine it with other Queries or specific column parsing
- DateRangeGenerator, which is a "column parser" useful if you have just ymd table and no other worries

Both have a similar API and behaviour:
- you specify the names of the ymd columns,
- you specify start and end as either string or datetime.date,
- the ranges are treated as [start, end).
"""

from __future__ import annotations

import calendar
import datetime
import logging
from enum import Enum, auto, unique
from typing import Optional, Union

from fsql import assert_exhaustive_enum
from fsql.column_parser import ColumnParser
from fsql.partition import Partition
from fsql.query import Query

logger = logging.getLogger(__name__)


def _date_parser(date_spec: Union[str, datetime.date]) -> datetime.date:
    if isinstance(date_spec, datetime.date):
        return date_spec
    else:
        return datetime.datetime.strptime(date_spec, "%Y/%m/%d").date()


class DateRangeQuery(Query):
    """Allows for selection ymd-partitions that fall into a given range."""

    def __init__(
        self,
        start: Union[str, datetime.date],
        end: Union[str, datetime.date],
        year_name: str = "year",
        month_name: str = "month",
        day_name: str = "day",
    ):
        """Evaluates to true for all ymd-partitions that are in the range [start, end). The range is specified either
        as "yyyy/mm/dd" string, or as date object. The `*_name` params are the target partition columns from which to
        construct the date object."""
        self.map = {"year": year_name, "month": month_name, "day": day_name}
        self.start = _date_parser(start)
        self.end = _date_parser(end)

    def eval_all(self, columns: dict[str, str]) -> bool:
        if not set(self.map.values()).issubset(columns.keys()):
            rv = False
        else:
            extractor = lambda s: columns[self.map[s]]  # noqa: E731
            extracted = f"{extractor('year')}/{extractor('month')}/{extractor('day')}"
            partition_value = _date_parser(extracted)
            rv = partition_value >= self.start and partition_value < self.end
        logger.debug(f"invoked daterange-all with {columns}, resulted to {rv}")
        return rv

    def eval_available(self, columns: dict[str, str]) -> bool:
        # here we need to eval that "this current partition can expand to at least one legit date"
        # we do this by generating the leftmost and rightmost dates, and comparing against end/start
        # in case all ymd are available already, this evaluates identically to eval_all
        if not self.map["year"] in columns:
            rv = True
        else:
            year = int(columns[self.map["year"]])
            month_l = int(columns.get(self.map["month"], "1"))
            month_r = int(columns.get(self.map["month"], "12"))
            day_l = int(columns.get(self.map["day"], "1"))
            day_r = int(columns.get(self.map["day"], str(calendar.monthrange(year, month_r)[1])))
            date_l = datetime.date(year, month_l, day_l)
            date_r = datetime.date(year, month_r, day_r)
            rv = date_l < self.end and date_r >= self.start
        logger.debug(f"invoked daterange-available with {columns}, resulted to {rv}")
        return rv


@unique
class DRLevel(Enum):
    Y = auto()
    M = auto()
    D = auto()
    F = auto()


class DateRangeGenerator(ColumnParser):
    """Use this in place of AutoParser or FixedColumnParser, if you have just ymd columns and want
    a date range partition. Can work with both =- and Fixed format. To determine which to use, see
    the `build` method -- which is the entrypoint you should use."""

    def __init__(
        self,
        start: datetime.date,
        end: datetime.date,
        level: DRLevel,
        map: dict[DRLevel, str],
        include_column_in_path: bool,
    ):
        self.start = start
        self.end = end
        self.level = level
        self.map = map
        self.include_column_in_path = include_column_in_path

    def __call__(self, dirname: str) -> tuple[str, str]:
        if self.include_column_in_path:
            key, value = dirname.strip("/").split("=", 1)
        else:
            key = self.map[self.level]
            value = dirname.strip("/")
        return key, value

    def tail(self, partition: Partition) -> ColumnParser:
        # this is heavy metal -- we need to restrict the start/end range to those valid for the current partition
        year = int(partition.columns[self.map[DRLevel.Y]])
        if self.level is DRLevel.Y:
            start = self.start if year == self.start.year else datetime.date(year, 1, 1)
            end = self.end if year == self.end.year else datetime.date(year, 12, 31)
            next_level = DRLevel.M
        elif self.level is DRLevel.M:
            month = int(partition.columns[self.map[DRLevel.M]])
            start = self.start if month == self.start.month else datetime.date(year, month, 1)
            month_range = calendar.monthrange(year, month)[1]
            end = self.end if month == self.end.month else datetime.date(year, month, month_range)
            next_level = DRLevel.D
        elif self.level is DRLevel.D:
            start = self.start
            end = self.end
            next_level = DRLevel.F
        else:
            raise ValueError("unexpected call of tail -- internal failure to terminate discovery")
        return DateRangeGenerator(start, end, next_level, self.map, self.include_column_in_path)

    def parses_filenames(self) -> bool:
        return False

    def is_terminal_level(self) -> bool:
        return self.level == DRLevel.F

    def generate(self) -> Optional[list[str]]:
        # we add the +1 because we need inclusive ranges
        if self.level is DRLevel.Y:
            int_range = range(self.start.year, self.end.year + 1)
        elif self.level is DRLevel.M:
            int_range = range(self.start.month, self.end.month + 1)
        elif self.level is DRLevel.D:
            int_range = range(self.start.day, self.end.day + 1)
        elif self.level is DRLevel.F:
            return None
        else:
            assert_exhaustive_enum(self.level)
        pref = f"{self.map[self.level]}=" if self.include_column_in_path else ""
        return [f"{pref}{str(e)}" for e in int_range]

    @classmethod
    def from_str(cls, path_description: str):
        raise NotImplementedError(f"class {cls} does not support this method")

    @classmethod
    def build(
        cls,
        start: Union[str, datetime.date],
        end: Union[str, datetime.date],
        year_name: str = "year",
        month_name: str = "month",
        day_name: str = "day",
        include_column_in_path: bool = True,
    ):
        """Generates all ymd-partitions that are in the range [start, end). The range is specified either
        as "yyyy/mm/dd" string, or as date object. The `*_name` params how should the resulting columns be called.
        If you set include_column_in_path, we generate partitions `{year_name}=2022` etc, otherwise just `2022` etc.
        """
        map = {DRLevel.Y: year_name, DRLevel.M: month_name, DRLevel.D: day_name}
        start = _date_parser(start)
        end = _date_parser(end) - datetime.timedelta(1)  # internally, we treat the date as inclusive!
        return cls(start, end, DRLevel.Y, map, include_column_in_path)
